@@ -50,12 +50,23 @@
 
 static int volatile registered = 0;
 
-static int always_show_icon = 1;
+static DBusGConnection *conn;
 
-static GtkStatusIcon *statusicon = NULL;
+typedef enum {
+	ICON_POLICY_NEVER,
+	ICON_POLICY_ALWAYS,
+	ICON_POLICY_PRESENT,
+} EnumIconPolicy;
+
+static int icon_policy = ICON_POLICY_PRESENT;
+
+#define PREF_DIR		"/apps/bluetooth-manager"
+#define PREF_ICON_POLICY	PREF_DIR "/icon_policy"
 
 static GConfClient* gconf;
 
+static GtkStatusIcon *statusicon = NULL;
+
 typedef enum {
 	PASSKEY_AGENT_ERROR_REJECT
 } PasskeyAgentError;
@@ -109,7 +120,21 @@ static gint input_compare(gconstpointer a, gconstpointer b)
 	return strcmp(a_data->address, b_data->address);
 }
 
-static void response_callback(GtkWidget *dialog,
+static void input_free(struct input_data *input)
+{
+	gtk_widget_destroy(input->dialog);
+
+	input_list = g_list_remove(input_list, input);
+
+	g_free(input->address);
+	g_free(input->path);
+	g_free(input);
+
+	if (g_list_length(input_list) == 0)
+		gtk_status_icon_set_blinking(statusicon, FALSE);
+}
+
+static void passkey_callback(GtkWidget *dialog,
 					gint response, gpointer user_data)
 {
 	struct input_data *input = user_data;
@@ -126,16 +151,20 @@ static void response_callback(GtkWidget *dialog,
 		dbus_g_method_return_error(input->context, error);
 	}
 
-	gtk_widget_destroy(dialog);
+	input_free(input);
+}
 
-	input_list = g_list_remove(input_list, input);
+static void confirm_callback(GtkWidget *dialog,
+					gint response, gpointer user_data)
+{
+	struct input_data *input = user_data;
+	gboolean result;
 
-	g_free(input->address);
-	g_free(input->path);
-	g_free(input);
+	result = (response == GTK_RESPONSE_YES) ? TRUE : FALSE;
 
-	if (g_list_length(input_list) == 0)
-		gtk_status_icon_set_blinking(statusicon, FALSE);
+	dbus_g_method_return(input->context, result);
+
+	input_free(input);
 }
 
 static void changed_callback(GtkWidget *editable, gpointer user_data)
@@ -158,8 +187,8 @@ static void toggled_callback(GtkWidget *button, gpointer user_data)
 	gtk_entry_set_visibility(GTK_ENTRY(input->entry), mode);
 }
 
-static void create_dialog(const char *path, const char *address,
-			gchar *device, DBusGMethodInvocation *context)
+static void passkey_dialog(const char *path, const char *address,
+			const gchar *device, DBusGMethodInvocation *context)
 {
 	GtkWidget *dialog;
 	GtkWidget *button;
@@ -279,7 +308,117 @@ static void create_dialog(const char *path, const char *address,
 	input_list = g_list_append(input_list, input);
 
 	g_signal_connect(G_OBJECT(dialog), "response",
-				G_CALLBACK(response_callback), input);
+				G_CALLBACK(passkey_callback), input);
+
+	gtk_status_icon_set_blinking(statusicon, TRUE);
+}
+
+static void confirm_dialog(const char *path, const char *address,
+				const char *value, const gchar *device,
+					DBusGMethodInvocation *context)
+{
+	GtkWidget *dialog;
+	GtkWidget *button;
+	GtkWidget *image;
+	GtkWidget *label;
+	GtkWidget *table;
+	gchar *markup;
+	struct input_data *input;
+
+	input = g_try_malloc0(sizeof(*input));
+	if (!input)
+		return;
+
+	input->path = g_strdup(path);
+	input->address = g_strdup(address);
+
+	input->context = context;
+
+	dialog = gtk_dialog_new();
+
+	gtk_window_set_title(GTK_WINDOW(dialog), _("Confirmation request"));
+
+	gtk_window_set_icon_name(GTK_WINDOW(dialog), "stock_bluetooth");
+
+	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);
+
+	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
+
+	gtk_window_set_keep_above(GTK_WINDOW(dialog), TRUE);
+
+	gtk_window_set_urgency_hint(GTK_WINDOW(dialog), TRUE);
+
+	gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
+
+	input->dialog = dialog;
+
+	button = gtk_dialog_add_button(GTK_DIALOG(dialog),
+					GTK_STOCK_NO, GTK_RESPONSE_NO);
+ 
+	button = gtk_dialog_add_button(GTK_DIALOG(dialog),
+					GTK_STOCK_YES, GTK_RESPONSE_YES);
+
+	table = gtk_table_new(4, 2, FALSE);
+
+	gtk_table_set_row_spacings(GTK_TABLE(table), 4);
+	gtk_table_set_col_spacings(GTK_TABLE(table), 20);
+
+	gtk_container_set_border_width(GTK_CONTAINER(table), 12);
+
+	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), table);
+
+	image = gtk_image_new_from_icon_name(GTK_STOCK_DIALOG_AUTHENTICATION,
+							GTK_ICON_SIZE_DIALOG);
+
+	gtk_misc_set_alignment(GTK_MISC(image), 0.0, 0.0);
+
+	gtk_table_attach(GTK_TABLE(table), image, 0, 1, 0, 5,
+						GTK_SHRINK, GTK_FILL, 0, 0);
+
+	label = gtk_label_new(_("Pairing request for device:"));
+
+	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
+
+	gtk_table_attach(GTK_TABLE(table), label, 1, 2, 0, 1,
+				GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
+
+	label = gtk_label_new(device);
+
+	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
+
+	gtk_label_set_selectable(GTK_LABEL(label), TRUE);
+
+	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
+
+	gtk_widget_set_size_request(GTK_WIDGET(label), 260, -1);
+
+	gtk_table_attach(GTK_TABLE(table), label, 1, 2, 1, 2,
+			GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 4);
+
+	label = gtk_label_new(_("\nConfirm value for authentication:"));
+
+	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
+
+	gtk_table_attach(GTK_TABLE(table), label, 1, 2, 2, 3,
+				GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
+
+	label = gtk_label_new(NULL);
+
+	markup = g_strdup_printf("<b>%s</b>\n", value);
+
+	gtk_label_set_markup(GTK_LABEL(label), markup);
+
+	g_free(markup);
+
+	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);
+
+	gtk_table_attach(GTK_TABLE(table), label, 1, 2, 3, 4,
+				GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);
+
+	input_list = g_list_append(input_list, input);
+
+	g_signal_connect(G_OBJECT(dialog), "response",
+				G_CALLBACK(confirm_callback), input);
 
 	gtk_status_icon_set_blinking(statusicon, TRUE);
 }
@@ -358,7 +497,6 @@ static void close_notification(void)
 
 typedef struct {
 	GObject parent;
-	DBusGConnection *conn;
 } PasskeyAgent;
 
 typedef struct {
@@ -376,10 +514,6 @@ G_DEFINE_TYPE(PasskeyAgent, passkey_agent, G_TYPE_OBJECT)
 
 static void passkey_agent_finalize(GObject *obj)
 {
-	PasskeyAgent *agent = PASSKEY_AGENT(obj);
-
-	dbus_g_connection_unref(agent->conn);
-
 	parent_class->finalize(obj);
 }
 
@@ -397,14 +531,12 @@ static void passkey_agent_class_init(PasskeyAgentClass *klass)
 	gobject_class->finalize = passkey_agent_finalize;
 }
 
-static PasskeyAgent *passkey_agent_new(DBusGConnection *conn, const char *path)
+static PasskeyAgent *passkey_agent_new(const char *path)
 {
 	PasskeyAgent *agent;
 
 	agent = g_object_new(PASSKEY_AGENT_OBJECT_TYPE, NULL);
 
-	agent->conn = dbus_g_connection_ref(conn);
-
 	dbus_g_connection_register_g_object(conn, path, G_OBJECT(agent));
 
 	return agent;
@@ -425,7 +557,7 @@ static gboolean passkey_agent_request(PasskeyAgent *agent,
 	const char *adapter = NULL, *name = NULL;
 	gchar *device, *text;
 
-	object = dbus_g_proxy_new_for_name(agent->conn, "org.bluez",
+	object = dbus_g_proxy_new_for_name(conn, "org.bluez",
 						path, "org.bluez.Adapter");
 
 	dbus_g_proxy_call(object, "GetName", NULL, G_TYPE_INVALID,
@@ -443,9 +575,51 @@ static gboolean passkey_agent_request(PasskeyAgent *agent,
 	} else
 		device = g_strdup(address);
 
-	create_dialog(path, address, device, context);
+	passkey_dialog(path, address, device, context);
 
-	text = g_strdup_printf(_("Pairing request for %s"), device);
+	text = g_strdup_printf(_("Pairing request for %s\n\n"
+			"Click to open passkey entry dialog"), device);
+
+	g_free(device);
+
+	show_notification(adapter ? adapter : _("Bluetooth device"),
+				text, 0, G_CALLBACK(notification_closed));
+
+	g_free(text);
+
+	return TRUE;
+}
+
+static gboolean passkey_agent_confirm(PasskeyAgent *agent,
+			const char *path, const char *address,
+			const char *value, DBusGMethodInvocation *context)
+{
+	DBusGProxy *object;
+	const char *adapter = NULL, *name = NULL;
+	gchar *device, *text;
+
+	object = dbus_g_proxy_new_for_name(conn, "org.bluez",
+						path, "org.bluez.Adapter");
+
+	dbus_g_proxy_call(object, "GetName", NULL, G_TYPE_INVALID,
+				G_TYPE_STRING, &adapter, G_TYPE_INVALID);
+
+	dbus_g_proxy_call(object, "GetRemoteName", NULL,
+				G_TYPE_STRING, address, G_TYPE_INVALID,
+				G_TYPE_STRING, &name, G_TYPE_INVALID);
+
+	if (name) {
+		if (g_strrstr(name, address))
+			device = g_strdup(name);
+		else
+			device = g_strdup_printf("%s (%s)", name, address);
+	} else
+		device = g_strdup(address);
+
+	confirm_dialog(path, address, value, device, context);
+
+	text = g_strdup_printf(_("Pairing request for %s\n\n"
+			"Click to open confirmation dialog"), device);
 
 	g_free(device);
 
@@ -461,6 +635,7 @@ static gboolean passkey_agent_cancel(PasskeyAgent *agent,
 			const char *path, const char *address, GError **error)
 {
 	GList *list;
+	GError *result;
 	struct input_data *input;
 
 	input = g_try_malloc0(sizeof(*input));
@@ -483,7 +658,12 @@ static gboolean passkey_agent_cancel(PasskeyAgent *agent,
 
 	close_notification();
 
-	response_callback(input->dialog, GTK_RESPONSE_DELETE_EVENT, input);
+	result = g_error_new(PASSKEY_AGENT_ERROR, PASSKEY_AGENT_ERROR_REJECT,
+						"Agent callback canceled");
+
+	dbus_g_method_return_error(input->context, result);
+
+	input_free(input);
 
 	return TRUE;
 }
@@ -497,7 +677,7 @@ static gboolean passkey_agent_release(PasskeyAgent *agent, GError **error)
 
 #include "passkey-agent-glue.h"
 
-static int register_passkey_agent(DBusGConnection *conn)
+static int register_passkey_agent(void)
 {
 	DBusGProxy *object;
 	GError *error = NULL;
@@ -510,8 +690,6 @@ static int register_passkey_agent(DBusGConnection *conn)
 				G_TYPE_INVALID, G_TYPE_INVALID);
 
 	if (error != NULL) {
-		g_printerr(_("Passkey agent registration failed: %s\n"),
-							error->message);
 		g_error_free(error);
 		return -1;
 	}
@@ -632,12 +810,30 @@ static gint adapter_compare(gconstpointer a, gconstpointer b)
 	return strcmp(adapter->path, path);
 }
 
-static void add_adapter(DBusGConnection *conn, const char *path)
+static void adapter_count(gpointer data, gpointer user_data)
+{
+	struct adapter_data *adapter = data;
+	int *count = user_data;
+
+	if (adapter->attached)
+		(*count)++;
+}
+
+static int attached_adapters(void)
+{
+	int count = 0;
+
+	g_list_foreach(adapter_list, adapter_count, &count);
+
+	return count;
+}
+
+static void add_adapter(const char *path)
 {
 	GList *list;
 	DBusGProxy *object;
 	struct adapter_data *adapter;
-	
+
 	list = g_list_find_custom(adapter_list, path, adapter_compare);
 	if (list && list->data) {
 		struct adapter_data *adapter = list->data;
@@ -651,6 +847,7 @@ static void add_adapter(DBusGConnection *conn, const char *path)
 		return;
 
 	adapter->path = g_strdup(path);
+	adapter->attached = 1;
 
 	adapter_list = g_list_append(adapter_list, adapter);
 
@@ -661,35 +858,31 @@ static void add_adapter(DBusGConnection *conn, const char *path)
 					G_TYPE_STRING, G_TYPE_INVALID);
 
 	dbus_g_proxy_connect_signal(object, "ModeChanged",
-				G_CALLBACK(mode_changed), conn, NULL);
+				G_CALLBACK(mode_changed), NULL, NULL);
 
 	dbus_g_proxy_add_signal(object, "BondingCreated",
 					G_TYPE_STRING, G_TYPE_INVALID);
 
 	dbus_g_proxy_connect_signal(object, "BondingCreated",
-				G_CALLBACK(bonding_created), conn, NULL);
+				G_CALLBACK(bonding_created), NULL, NULL);
 
 	dbus_g_proxy_add_signal(object, "BondingRemoved",
 					G_TYPE_STRING, G_TYPE_INVALID);
 
 	dbus_g_proxy_connect_signal(object, "BondingRemoved",
-				G_CALLBACK(bonding_removed), conn, NULL);
+				G_CALLBACK(bonding_removed), NULL, NULL);
 }
 
 static void adapter_added(DBusGProxy *object,
 				const char *path, gpointer user_data)
 {
-	DBusGConnection *conn = user_data;
-
 	if (!registered)
-		register_passkey_agent(conn);
+		register_passkey_agent();
 
-	add_adapter(conn, path);
+	add_adapter(path);
 
-	if (!always_show_icon) {
-		if (g_list_length(adapter_list) > 0)
-			gtk_status_icon_set_visible(statusicon, TRUE);
-	}
+	if (icon_policy != ICON_POLICY_NEVER && attached_adapters() > 0)
+		gtk_status_icon_set_visible(statusicon, TRUE);
 }
 
 static void adapter_removed(DBusGProxy *object,
@@ -704,13 +897,11 @@ static void adapter_removed(DBusGProxy *object,
 		adapter->attached = 0;
 	}
 
-	if (!always_show_icon) {
-		if (g_list_length(adapter_list) == 0)
-			gtk_status_icon_set_visible(statusicon, FALSE);
-	}
+	if (icon_policy != ICON_POLICY_ALWAYS && attached_adapters() == 0)
+		gtk_status_icon_set_visible(statusicon, FALSE);
 }
 
-static int setup_manager(DBusGConnection *conn)
+static int setup_manager(void)
 {
 	DBusGProxy *object;
 	GError *error = NULL;
@@ -723,27 +914,24 @@ static int setup_manager(DBusGConnection *conn)
 					G_TYPE_STRING, G_TYPE_INVALID);
 
 	dbus_g_proxy_connect_signal(object, "AdapterAdded",
-				G_CALLBACK(adapter_added), conn, NULL);
+				G_CALLBACK(adapter_added), NULL, NULL);
 
 	dbus_g_proxy_add_signal(object, "AdapterRemoved",
 					G_TYPE_STRING, G_TYPE_INVALID);
 
 	dbus_g_proxy_connect_signal(object, "AdapterRemoved",
-				G_CALLBACK(adapter_removed), conn, NULL);
+				G_CALLBACK(adapter_removed), NULL, NULL);
 
 	dbus_g_proxy_call(object, "ListAdapters", &error,
 			G_TYPE_INVALID,	G_TYPE_STRV, &array, G_TYPE_INVALID);
 
-	if (error != NULL) {
-		g_printerr(_("Getting adapter list failed: %s\n"),
-							error->message);
-		g_error_free(error);
-	} else {
+	if (error == NULL) {
 		while (*array) {
-			add_adapter(conn, *array);
+			add_adapter(*array);
 			array++;
 		}
-	}
+	} else
+		g_error_free(error);
 
 	return 0;
 }
@@ -751,17 +939,17 @@ static int setup_manager(DBusGConnection *conn)
 static void name_owner_changed(DBusGProxy *object, const char *name,
 			const char *prev, const char *new, gpointer user_data)
 {
-	if (strcmp(name, "org.bluez") && *new == '\0') {
+	if (!strcmp(name, "org.bluez") && *new == '\0') {
 		registered = 0;
 
 		g_list_foreach(adapter_list, adapter_disable, NULL);
 
-		if (!always_show_icon)
+		if (icon_policy != ICON_POLICY_ALWAYS)
 			gtk_status_icon_set_visible(statusicon, FALSE);
 	}
 }
 
-static int setup_dbus(DBusGConnection *conn)
+static int setup_dbus(void)
 {
 	DBusGProxy *object;
 	void *agent;
@@ -781,7 +969,7 @@ static int setup_dbus(DBusGConnection *conn)
 	dbus_g_error_domain_register(PASSKEY_AGENT_ERROR, "org.bluez.Error",
 						PASSKEY_AGENT_ERROR_TYPE);
 
-	agent = passkey_agent_new(conn, AGENT_PATH);
+	agent = passkey_agent_new(AGENT_PATH);
 
 	return 0;
 }
@@ -831,7 +1019,15 @@ static void about_callback(GtkWidget *item, gpointer user_data)
 	gtk_widget_show_all(dialog);
 }
 
-static void activate_callback(GObject *object, gpointer user_data)
+static void settings_callback(GObject *widget, gpointer user_data)
+{
+        const char *command = "bluetooth-properties";
+
+        if (!g_spawn_command_line_async(command, NULL))
+		g_printerr(_("Couldn't execute command: %s\n"), command);
+}
+
+static void activate_callback(GObject *widget, gpointer user_data)
 {
 	close_notification();
 
@@ -840,14 +1036,34 @@ static void activate_callback(GObject *object, gpointer user_data)
 	gtk_status_icon_set_blinking(statusicon, FALSE);
 }
 
-static void popup_callback(GObject *object, guint button,
+static void popup_callback(GObject *widget, guint button,
 				guint activate_time, gpointer user_data)
 {
+	DBusGProxy *object;
 	GtkWidget *menu;
 	GtkWidget *item;
+	const char *mode;
 
 	menu = gtk_menu_new();
 
+	object = dbus_g_proxy_new_for_name(conn, "org.bluez",
+				"/org/bluez/hci0", "org.bluez.Adapter");
+
+	dbus_g_proxy_call(object, "GetMode", NULL, G_TYPE_INVALID,
+				G_TYPE_STRING, &mode, G_TYPE_INVALID);
+
+#if 0
+	item = gtk_separator_menu_item_new();
+	gtk_widget_show(item);
+	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
+#endif
+
+	item = gtk_image_menu_item_new_from_stock(GTK_STOCK_PREFERENCES, NULL);
+	g_signal_connect(item, "activate",
+				G_CALLBACK(settings_callback), NULL);
+	gtk_widget_show(item);
+	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
+
 	item = gtk_image_menu_item_new_from_stock(GTK_STOCK_ABOUT, NULL);
 	g_signal_connect(item, "activate",
 				G_CALLBACK(about_callback), NULL);
@@ -856,18 +1072,50 @@ static void popup_callback(GObject *object, guint button,
 
 	gtk_menu_popup(GTK_MENU(menu), NULL, NULL,
 			gtk_status_icon_position_menu,
-			GTK_STATUS_ICON(object), button, activate_time);
+			GTK_STATUS_ICON(widget), button, activate_time);
 }
 
+static GConfEnumStringPair icon_policy_enum_map [] = {
+	{ ICON_POLICY_NEVER,	"never"		},
+	{ ICON_POLICY_ALWAYS,	"always"	},
+	{ ICON_POLICY_PRESENT,	"present"	},
+	{ ICON_POLICY_PRESENT,	NULL		},
+};
+
 static void gconf_callback(GConfClient *client, guint cnxn_id,
 					GConfEntry *entry, gpointer user_data)
 {
+        if (gconf_entry_get_value(entry) == NULL)
+                return;
+
+        if (strcmp(entry->key, PREF_ICON_POLICY) == 0) {
+		gboolean visible;
+		char *str;
+
+		str = gconf_client_get_string(client, entry->key, NULL);
+		if (str) {
+		        gconf_string_to_enum(icon_policy_enum_map,
+							str, &icon_policy);
+			g_free(str);
+		}
+
+		visible = gtk_status_icon_get_visible(statusicon);
+
+		if (icon_policy == ICON_POLICY_NEVER)
+			visible = FALSE;
+		else if (icon_policy == ICON_POLICY_ALWAYS)
+			visible = TRUE;
+		else if (icon_policy == ICON_POLICY_PRESENT)
+			visible = attached_adapters() > 0 ? TRUE : FALSE;
+
+		gtk_status_icon_set_visible(statusicon, visible);
+	}
 }
 
 int main(int argc, char *argv[])
 {
-	DBusGConnection *conn;
 	GError *error = NULL;
+	char *str;
 
 	bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
 	bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
@@ -876,7 +1124,7 @@ int main(int argc, char *argv[])
 	gtk_init(&argc, &argv);
 
 #ifdef HAVE_LIBNOTIFY
-	notify_init("bt-applet");
+	notify_init("bluetooth-mananger");
 #endif
 
 	conn = dbus_g_bus_get(DBUS_BUS_SYSTEM, &error);
@@ -889,10 +1137,15 @@ int main(int argc, char *argv[])
 
 	gconf = gconf_client_get_default();
 
-	gconf_client_add_dir(gconf, "/apps/bt-applet",
-					GCONF_CLIENT_PRELOAD_NONE, NULL);
+	str = gconf_client_get_string(gconf, PREF_ICON_POLICY, NULL);
+	if (str) {
+	        gconf_string_to_enum(icon_policy_enum_map, str, &icon_policy);
+	        g_free(str);
+	}
 
-	gconf_client_notify_add(gconf, "/apps/bt-applet",
+	gconf_client_add_dir(gconf, PREF_DIR, GCONF_CLIENT_PRELOAD_NONE, NULL);
+
+	gconf_client_notify_add(gconf, PREF_DIR,
 					gconf_callback, NULL, NULL, NULL);
 
 	statusicon = gtk_status_icon_new_from_icon_name("stock_bluetooth");
@@ -903,18 +1156,16 @@ int main(int argc, char *argv[])
 				G_CALLBACK(activate_callback), NULL);
 
 	g_signal_connect(statusicon, "popup-menu",
-				G_CALLBACK(popup_callback), NULL);
+				G_CALLBACK(popup_callback), conn);
 
-	setup_dbus(conn);
+	setup_dbus();
 
-	setup_manager(conn);
+	setup_manager();
 
-	if (!always_show_icon) {
-		gtk_status_icon_set_visible(statusicon,
-			g_list_length(adapter_list) > 0 ? TRUE : FALSE);
-	}
+	if (icon_policy == ICON_POLICY_NEVER || attached_adapters() == 0)
+		gtk_status_icon_set_visible(statusicon, FALSE);
 
-	register_passkey_agent(conn);
+	register_passkey_agent();
 
 	gtk_main();
 
@@ -924,5 +1175,7 @@ int main(int argc, char *argv[])
 
 	g_list_foreach(adapter_list, adapter_free, NULL);
 
+	dbus_g_connection_unref(conn);
+
 	return 0;
 }
